/*
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.  
 *
 * * Neither the name of the University of California, Berkeley nor
 *   the names of its contributors may be used to endorse or promote
 *   products derived from this software without specific prior 
 *   written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package blog;

import java.util.*;

/*
 * Java implementation of a BLOG non-guaranteed object. A BLOG
 * non-guaranteed object is uniquely specified by a POP and associated
 * generating objects that generated this object, and a number
 * specifying its position in the generative process.
 */
public class NonGuaranteedObject {
    /*
     * @param pop             The POP that generated this object.
     * @param genObjs         The generating objects used with this POP.
     * @param numb            This object is the i'th generated from
     *                          the above mentiond POP/generating objects in
     *                          this world.
     */
    public NonGuaranteedObject(POP pop, List genObjs, int numb) {
	this.pop = pop;
	this.genObjs = genObjs;
	this.numb = numb;
    }

    /*
     * @param generatedFrom   The NumberVar object that uniquely
     *                          specifies the POP and generating objects
     *                          that generated this object.
     * @param numb            This object is the i'th generated from
     *                          the above mentiond POP/generating objects in
     *                          this world.
     */
    public NonGuaranteedObject(NumberVar generatedFrom, int numb) {
	this.pop = generatedFrom.pop();
	this.genObjs = generatedFrom.args();
	this.numb = numb;
    }

    public int getNumber() {
	return numb;
    }

    public POP getPOP() {
	return pop;
    }

    public List getGeneratingObjs() {
	return genObjs;
    }

    /**
     * Returns the value of the given generating function on this object.
     */
    public Object getGenFuncValue(GeneratingFunction genFunc) {
	int index = pop.getGenFuncIndex(genFunc);
	if (index >= 0) {
	    return genObjs.get(index);
	}

	// genFunc is not used in POP that generated this obj
	return Model.NULL; 
    }

    /**
     * Returns the BLOG type of this non-guaranteed object.
     */
    public Type getType() {
	return pop.getType();
    }

    /**
     * Returns the number variable whose value determines whether this 
     * non-guaranteed object exists or not.
     */
    public NumberVar getNumberVar() {
	return new NumberVar(pop, genObjs);
    }

    /**
     * Returns the depth of nesting in this tuple.  A guaranteed object or 
     * an object generated by the empty tuple has depth zero.  The depth of 
     * any other object is one more than the maximum depth of its generating 
     * objects.
     */
    public int getDepth() {
	if (genObjs.isEmpty()) {
	    return 0;
	}

	int maxGenObjDepth = 0;
	for (Iterator iter = genObjs.iterator(); iter.hasNext(); ) {
	    Object genObj = iter.next();
	    if (genObj instanceof NonGuaranteedObject) {
		int d = ((NonGuaranteedObject) genObj).getDepth();
		maxGenObjDepth = Math.max(d, maxGenObjDepth);
	    } // else genObj is a guaranteed object, depth 0
	}
	
	return maxGenObjDepth + 1;
    }
    
    public boolean equals(Object obj) {
	if(obj != null && obj instanceof NonGuaranteedObject) {
	    NonGuaranteedObject other = (NonGuaranteedObject) obj;

	    return (getNumber() == other.getNumber()) &&
		(getPOP().equals(other.getPOP())) &&
		(getGeneratingObjs()).equals(other.getGeneratingObjs());
	}
	return false;
    }

    public int hashCode() {
	return numb ^ getPOP().hashCode() ^ getGeneratingObjs().hashCode();
    }

    /**
     * Returns the tuple representation of this object.
     */
    public String toString() {
	StringBuffer buf = new StringBuffer();
	buf.append("(" + getType() + ", ");
	
	for (int i = 0; i < genObjs.size(); ++i) {
	    buf.append("(" + pop.getGeneratingFunctions().get(i) 
		       + ", " + genObjs.get(i) + "), ");
	}

	buf.append(numb + ")");
	return buf.toString();
    }

    /**
     * Compares this non-guaranteed object to another one.  The objects 
     * are first compared by depth: shallower objects come first.  Objects 
     * of the same depth are compared lexicographically according to their 
     * lists of generating objects.  Objects with the same generating 
     * objects but different POPs are compared according to the order in 
     * which their POPs were defined.  Finally, objects from the same POP 
     * application are compared by their indices.  
     */
    public int compareTo(Object o) {
	NonGuaranteedObject other = (NonGuaranteedObject) o;
	int depthDiff = getDepth() - other.getDepth();
	if (depthDiff != 0) {
	    return depthDiff;
	}

	int genObjListDiff = Model.compareArgLists(genObjs, 
						   other.getGeneratingObjs());
	if (genObjListDiff != 0) {
	    return genObjListDiff;
	}

	int popDiff = pop.getDepModel().getCreationIndex() 
	    - other.getPOP().getDepModel().getCreationIndex();
	if (popDiff != 0) {
	    return popDiff;
	}

	return (numb - other.getNumber());
    }

    private POP pop;
    private List genObjs;
    private int numb;
}
