/*
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.  
 *
 * * Neither the name of the University of California, Berkeley nor
 *   the names of its contributors may be used to endorse or promote
 *   products derived from this software without specific prior 
 *   written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package blog;

import java.util.*;
import java.io.PrintStream;

import common.DGraph;

/**
 * A partial description of a possible world in a BLOG model.  This 
 * description stands for the set of all possible worlds that satisfy the 
 * description.  More specifically, a PartialWorld consists of a set of 
 * object identifiers and an instantiation of a certain set of random 
 * variables.  These random variables can be basic variables of the BLOG 
 * model (number variables and random function application variables) 
 * parameterized by objects or identifiers.  They can also be generating 
 * function application variables parameterized by identifiers 
 * (the values of generating functions on actual objects (non-identifiers) 
 * are fully determined).  
 *
 * <p>A PartialWorld stands for the set of possible worlds such that for some 
 * assignment of distinct non-guaranteed objects to the object identifiers, 
 * all the instantiated variables have the values specified by the 
 * PartialWorld.  
 *
 * <p>The value of the number variable for a POP application may be
 * stored separately from the set of identifiers that satisfy that POP
 * application.  Thus, some implementations of PartialWorld may allow
 * you to create a description that corresponds to an empty set of
 * worlds: one where the number of identifiers satisfying a POP
 * application is greater than the number of objects generated by that
 * POP application.  It may also be possible to create a PartialWorld
 * where some identifiers do not satisfy any POP application.
 */
public interface PartialWorld {
    /**
     * A partial world that does not assign values to any basic random 
     * variables.
     */
    static final PartialWorld EMPTY_INST 
	= new AbstractPartialWorld(Collections.EMPTY_SET) {
		protected Map getVarValues() {
		    return Collections.EMPTY_MAP;
		}

		public DGraph getBayesNet() {
		    return DGraph.EMPTY_GRAPH;
		}
	    };

    /**
     * Returns the set of object identifiers used in this PartialWorld.
     *
     * @return unmodifiable Set of ObjectIdentifier objects
     */
    Set getIdentifiers();

    /**
     * Returns true if the given variable is instantiated in this 
     * partial world.  If the given variable refers to objects that do 
     * not necessarily exist in this partial world, the method just 
     * returns false.
     */
    boolean isInstantiated(BayesNetVar var);

    /**
     * Returns an unmodifiable Set representing the set of variables that 
     * are instantiated in this world.
     *
     * @return unmodifiable Set of BayesNetVar objects
     */
    Set getInstantiatedVars();

    /**
     * Returns the value of the given variable in this partial world.  
     *
     * @throws IllegalArgumentException if the given variable is not 
     *                                  instantiated in this world
     */
    Object getValue(BayesNetVar var);

    /**
     * Returns the value of the given function applied to the given 
     * tuple of objects in this world.  Returns null if some of the 
     * given arguments do not exist in this world.  
     *
     * @throws IllegalStateException    if this PartialWorld is not complete 
     *                                  enough to specify the desired value
     */
    Object getFuncValue(Function func, List args);

    /**
     * Returns the number of objects that satisfy the given POP
     * application in this world.  This is the value of the
     * corresponding number variable, not the number of identifiers in
     * the world that satisfy this POP.  If some of the given
     * generating objects do not exist, then the return value is zero.
     *
     * @param pop               a potential object pattern
     * @param generatingObjects a List of Objects
     *
     * @throws IllegalStateException if all the given generating objects 
     *                               exist in this world, but the 
     *                               corresponding number variable is 
     *                               not instantiated
     */
    int getNumSatisfiers(POP pop, List generatingObjects);

    /**
     * Returns the set of objects that satisfy the given POP in this
     * world.  The objects may be represented as actual objects or
     * identifiers, depending on the way this PartialWorld
     * implementation handles objects of the relevant type.  If the
     * objects are represented as identifiers and the number of
     * identifiers in the world that satisfy the POP is less than the
     * value of the corresponding number variable, then the returned 
     * set contains an IdentifierHook object representing the remaining 
     * objects.  
     *
     * <p>The set returned by this method will remain correct if new 
     * identifiers are added or new random variables are instantiated.  
     * It will not remain correct if identifiers are removed or 
     * already-instantiated random variables are changed.  
     *
     * @param pop                a potential object pattern
     * @param generatingObjects  a List of Objects
     *
     * @throws IllegalStateException if all the given generating objects 
     *                               exist in the world, but the corresponding 
     *                               number variable is not instantiated
     *
     * @throws IllegalStateException if there are some identifiers of the 
     *                               type generated by this POP that do 
     *                               not have all their generating functions 
     *                               instantiated, and thus may or may not 
     *                               satisfy this POP application
     */
    Set getSatisfiers(POP pop, List generatingObjects);

    /**
     * Returns the set of identifiers that satisfy the given POP application 
     * in this world.  This is an empty set if some of the given generating 
     * objects do not exist, or if this partial world does not use 
     * identifiers for the relevant type of object.  
     *
     * @param pop               a potential object pattern
     * @param generatingObjects a List of Objects
     *
     * @return unmodifiable Set of ObjectIdentifier
     *
     * @throws IllegalStateException if there are some identifiers of the 
     *                               type generated by this POP that do 
     *                               not have all their generating functions 
     *                               instantiated, and thus may or may not 
     *                               satisfy this POP application
     */
    Set getSatisfyingIds(POP pop, List generatingObjects);

    /**
     * Returns the number of objects that satisfy the given POP application 
     * in this world but are not represented by identifiers in this world.  
     * More precisely, the number returned is max(0, numsat - numid), 
     * where numsat is the value returned by getNumSatisfiers and 
     * numid is the size of the set returned by getSatisfyingIds.  
     *
     * <p>Note that by this definition, if the given POP application is for 
     * a type that is not represented with identifiers, then the return 
     * value is just numsat.
     *
     * @param pop                a potential object pattern
     * @param generatingObjects  a List of Objects
     *
     * @throws IllegalStateException if all the given generating objects 
     *                               exist in the world, but the corresponding 
     *                               number variable is not instantiated
     *
     * @throws IllegalStateException if there are some identifiers of the 
     *                               type generated by this POP that do 
     *                               not have all their generating functions 
     *                               instantiated, and thus may or may not 
     *                               satisfy this POP application
     */
    int getNumSatisfiersWithoutIds(POP pop, List generatingObjects);

    /**
     * Returns true if this PartialWorld determines which POP application 
     * (if any) the given object satisfies.
     *
     * @throws IllegalArgumentException if the given object does not exist 
     *                                  in this world
     */
    boolean isPOPAppDetermined(Object obj);

    /**
     * Returns the NumberVar (i.e., POP and generating objects) such that 
     * the given object satisfies that POP applied to those generating 
     * objects in this world.  Returns null if the given object does not 
     * satisfy any POP application.  
     *
     * @throws IllegalArgumentException if the given object does not exist 
     *                                  in this world
     *
     * @throws IllegalStateException    if the given object is an identifier 
     *                                  and this PartialWorld does not 
     *                                  specify the values of generating 
     *                                  functions on it 
     */
    NumberVar getPOPAppSatisfied(Object obj);

    /**
     * Returns true if <code>obj</code> is in this world's set of object 
     * identifiers or is an object that exists in all worlds consistent 
     * with this PartialWorld.
     */
    boolean isValidObject(Object obj);

    /**
     * Returns true if <code>isValidObject</code> would return true for 
     * every object in the given list.  
     *
     * @param objs a List of arbitrary objects
     */
    boolean areValidObjects(List objs);

    /**
     * Returns a directed graph where the nodes are variables, and
     * there is an edge from variable X to variable Y if Y depends
     * directly on X given the context represented by this partial
     * world.  The graph includes at least the variables that are
     * instantiated in this partial world.  It may also include
     * certain derived variables that are not instantiated because
     * their values are determined by the instantiated variables.
     */
    DGraph getBayesNet();

    /**
     * Returns the set of variables that have the given object as their 
     * value in this world.  This is done efficiently if the object is an 
     * ObjectIdentifier, and slowly (by iterating over all instantiated 
     * variables) otherwise.
     *
     * @return unmodifiable Set of BayesNetVar
     */
    Set getVarsWithValue(Object value);

    /**
     * Returns the set of instantiated variables that have the given
     * object as an argument.  This is done efficiently if the object
     * is an ObjectIdentifier, and slowly (by iterating over all
     * instantiated variables) otherwise.
     *
     * @return unmodifiable Set of BayesNetVar
     */
    Set getVarsWithArg(Object arg);

    /**
     * Prints this partial world to the given stream.
     */
    void print(PrintStream s);

    /**
     * Adds the given listener to a set of weak references to objects 
     * that should be notified when an identifier begins or ceases 
     * to satisfy the given POP application.  
     *
     * <p>Currently there is no removeIdentifierListener method; a 
     * listener is removed from the listener set automatically when it 
     * is garbage-collected.  
     */ 
    void addIdentifierListener(POP pop, List genObjs, 
			       IdentifierListener listener);

    /**
     * Nested interface for objects that can be notified when an identifier 
     * is added or removed from a set.
     */
    static interface IdentifierListener {
	/**
	 * Called when the given identifier has just been added to the 
	 * set being monitored.
	 */
	void identifierAdded(ObjectIdentifier id);

	/**
	 * Called when the given identifier has just been removed from 
	 * the set being monitored.
	 */
	void identifierRemoved(ObjectIdentifier id);
    }

}
